;;; Copyright (C) 1995 Sho-Huan Simon Tung
;;; See ./VES2/COPYING for details

(define exist-in-top-level? #f)
(define add-top-level #f)
(define assign-top-level #f)
(define get-top-level #f)
(define env-init #f)
(define env-reset #f)
(define set-new&set #f)
(define GC #f)
(define lookup #f)
(define assign #f)
(define new-env #f)
(define get-max-frame-num #f)
(define get-num-of-frames #f)
(define get-used-frames #f)
(define get-track-parent #f)

(let ([num-of-frames 10000]
      [user-defined-top-level '(0)]
      [new-frames '()]
      [set-frames '()]
      [used-frames '*]
      [track-parent '*]
      [mark-frames '*]
      [max-frame-num 0])

(define user-defined-top-level
  (lambda ()
    (vector-ref used-frames 0)))

(define local-exist-in-top-level?
  (lambda (id)
    (find id (list (user-defined-top-level)))))

(define local-add-top-level
  (lambda (id val)
    (vector-set! used-frames 0
      (cons 0
	    (cons (list id val)
		  (cdr (user-defined-top-level)))))))

(define local-assign-top-level
  (lambda (id val)
    (local-assign id val (list (user-defined-top-level)))))

(define local-get-top-level
  (lambda ()
    (user-defined-top-level)))

(define local-env-init
  (lambda ()
    (set! new-frames '())
    (set! set-frames '())
    (set! used-frames (make-vector num-of-frames 0))
    (set! track-parent (make-vector num-of-frames -1))
    (vector-set! used-frames 0 (list 0))
    (set! mark-frames (make-vector num-of-frames 0))
    (set! max-frame-num 0)))

(define local-env-reset
  (lambda ()
    (set! new-frames '())
    (set! set-frames '())
    (vector-set! used-frames 0 (list 0))
    (let ([free (GC (user-defined-top-level) '())])
      (if (not (null? free))
	  (for-each (lambda (l)
		      (if display-env?
			  (begin
			    (pretty-print (list '!C l))
			    (kill-time 100000))))
	    free)))
    (if display-env?
	(pretty-print (list '!C 0)))
    (set! track-parent (make-vector num-of-frames -1))
    (set! used-frames (make-vector num-of-frames 0))
    (vector-set! used-frames 0 (list 0))
    (set! mark-frames (make-vector num-of-frames 0))
    (set! max-frame-num 0)))

(define local-set-new&set
  (lambda ()
    (set! new-frames '())
    (set! set-frames '())))
      
(define find-next-avail
  (lambda ()
    (call/cc
      (lambda (k)
	(if (>= max-frame-num (- num-of-frames 1))
	    (let ([min 	    ((rec loop
	       (lambda (n)
		 (if (eq? (vector-ref used-frames n) 0)
		     (begin
		       (vector-set! used-frames n '(n))
		       n)
		     (if (= n (- num-of-frames 1))
			 (error 'find-next-avail "No more frames.")
			 (loop (+ n 1)))))) 0)])
 	          (pretty-print
		   "WARNING: New frame number smaller than old frame numbers.")
		  min)
	    ((rec loop
	       (lambda (n)
		 (if (eq? (vector-ref used-frames n) 0)
		     (begin
		       (vector-set! used-frames n '(n))
		       (if (> n max-frame-num) (set! max-frame-num n))
		       n)
		     (loop (+ n 1))))) (+ 1 max-frame-num)))))))

(define mark-frame-no
  (lambda (env)
    (if (null? env)
	'()
	(begin
	  (vector-set! mark-frames (caar env) 1)
	  (mark-frame-no (cdr env))))))

(define mark-used-frames
  (lambda (exp-env)
    ((rec output-frm
       (lambda (rsl)
	   (if (pair? rsl)
	       (case (car rsl)
		 [(set!)
		  (if (exp-env? (cadr rsl))
		      (let ([v  (caddr rsl)])
			(output-frm v))
		      (map output-frm rsl))]
		 [(*closure*)
		  (let ([lam (cadr rsl)])
	            (mark-frame-no (caddr rsl)))]
		 [else (map output-frm rsl)])
	       (if (exp-env? rsl)
		   (let ([exp (exp-env-exp rsl)]
			 [env (exp-env-env rsl)])
		     (if (value? env)
			 (if (and (pair? exp) (eq? '*closure* (car exp)))
			     (mark-frame-no (caddr exp))
			     (if (eq? 'cont-obj env)
				 (output-frm (cadr exp))
				 '()))
			 (if (and (pair? exp) (not (eq? '*closure* (car exp))))
			     (begin
			       (map output-frm exp) 
			       (mark-frame-no env))
			     (mark-frame-no env))))
		   '())))) exp-env)))

(define mark-loop	; mark used environment from top-env.
  (lambda (top-env)
    (if (null? top-env)
	'()
	(let ([bind (car top-env)])
	  (let ([id (car bind)]
		[val (cadr bind)])
	    (if (and (pair? val) (eq? '*closure* (car val)))
		(let ([env (caddr val)])
		  ((rec loop
			(lambda (e)
			  (if (null? e)
			      '()
			      (if (= 1 (vector-ref mark-frames
						   (caar e)))
				  (loop (cdr e))
				  (begin
				    (vector-set!
				     mark-frames (caar e) 1)
				    (mark-loop (cdr (car e)))
				    (loop (cdr e))))))) env)
		  (mark-loop (cdr top-env)))
		(if (and (pair? val) (eq? '!K (car val)))
		    (let ([exp (cadr val)])
		      ((rec loop
			    (lambda (e)
			      (if (pair? e)
				  (map loop e)
				  (if (and (exp-env? e)
					   (not (value?
						 (exp-env-env e))))
				      (let ([env (exp-env-env e)])
					(map (lambda (i)
					       (if (not
						    (= 1
						       (vector-ref
							mark-frames
							(car i))))
						   (begin
						     (vector-set!
						      mark-frames
						      (car i) 1)
						     (mark-loop
						      (cdr i)))))
					     env)))))) exp)
		      (mark-loop (cdr top-env)))
		    (mark-loop (cdr top-env)))))))))

(define local-GC
  (lambda (env exp-env)
    (if #t
        ;(or (eq? #f current-exp-env)
	;    (and (exp-env? current-exp-env)
	;	 (value? (exp-env-env current-exp-env))))
	(begin
	  (vector-set! mark-frames 0 1)
          (mark-used-frames exp-env)
          (mark-loop (cdr env))
	  ((rec loop
		(lambda (n)
		  (if (> n 0)
		      (begin
			(if (= 1 (vector-ref mark-frames n))
			    (mark-loop (cdr (vector-ref used-frames n))))
			(loop (- n 1)))))) max-frame-num)
	  (let ([free '()])
	    ((rec loop
	       (lambda (n)
		 (if (> n max-frame-num)
		     '()
		     (if (and (= (vector-ref mark-frames n) 0)
			      (not (eq? (vector-ref used-frames n) 0)))
			 (begin
			   (set! free (cons n free))
			   (vector-set! used-frames n 0)
			   (vector-set! track-parent n -1)
			   (loop (+ n 1)))
			 (loop (+ n 1)))))) 1)
	    (set! mark-frames (make-vector num-of-frames 0))
	    (vector-set! mark-frames 0 1)
	    ((rec find-max (lambda (i)
				 (if (< i num-of-frames)
				     (begin 
				       (if (not (eq? 0
						 (vector-ref used-frames i)))
					   (set! max-frame-num i))
				       (find-max (+ i 1)))))) 0)
	    free))
	(begin
	  ; (printf "Cannot collect garbage while evaluating.~%")
	  '()))))

(define local-new-env
  (lambda (idl vals env tail fno)
    (if (and tail (not (= 0 fno)))
	(begin
	  (vector-set! used-frames fno 0)
	  (vector-set! track-parent fno -1)
	  (if display-env? (pretty-print (list '!C fno)))))
    (let ([parent-num (if (null? env) 0 (caar env))]
	  [new (new-env-help idl vals env)]
	  [v-frame-no (find-next-avail)])
      (set-car! new (cons v-frame-no (car new)))
      (vector-set! used-frames v-frame-no (car new))
      (vector-set! track-parent v-frame-no parent-num)
      (set! new-frames (cons (car new) new-frames))
      (for-each
	(lambda (f)
	  (if (> (length f) 0)
	      (if display-env?
		  (begin		    
		    (pretty-env
		      (list '!N (output-mgr
				  (cons (car f)
					  (cons parent-num (cdr f))))))
;;;   		    (list '!N parent-num (output-mgr f))
		    (newline)))))
	new-frames)
      (set! new-frames '())
      new)))
	    
(define new-env-help
  (lambda (idl vals env)
    (letrec ([make-frame
	       (lambda (idl vals)
		     (cond
		       [(null? idl) '()]
		       [else
			 (cons (list (car idl) (car vals))
			       (make-frame (cdr idl) (cdr vals)))]))]
	     [make-id-frame
	       (lambda (idl vals)
		 (list (list idl vals)))]
	     [make-pair-frame
	       (lambda (idl vals)
		 (cond
		   [(symbol? (cdr idl))
		    (cons (list (car idl) (car vals))
			  (cons (list (cdr idl) (cdr vals)) '()))]
		   [else
		     (cons (list (car idl) (car vals))
			   (make-pair-frame (cdr idl) (cdr vals)))]))])
      (cons
	(if (list? idl)
	    (make-frame idl vals)
	    (if (symbol? idl)
		(make-id-frame idl vals)
		(make-pair-frame idl vals)))
	env))))

(define find
  (lambda (id env)
    (if (null? env)
	#f
	(let ([num-frame (car env)])
	  (if (integer? (car num-frame))
	      (let ([frame (cdr num-frame)])
		(let ([pair (assq id frame)])
		  (if (not pair)
		      (find id (cdr env))
		      pair)))
	      (error 'find "Bad env"))))))

(define find-int
  (lambda (id env)
    (if (null? env)
	(error 'find-int "Wrong")
	(let ([num-frame (car env)])
	  (if (integer? (car num-frame))
	      (let ([frame (cdr num-frame)])
		(let ([pair (assq id frame)])
		  (if (not pair)
		      (find-int id (cdr env))
		      (car num-frame))))
	      (error 'find-int "Bad env"))))))

(define local-lookup
  (lambda (id env)
    (let ([pair (find id env)])
      (if (not pair)
	  (let ([pair0 (assq id (cdr (user-defined-top-level)))])
	    (if (not pair0)
		(top-level-value id)
		(cadr pair0)))
	  (let ([v (cadr pair)])
	    (if (exp-env? v)
		(error 'lookup "Wrong value in env ~s" v)
		v))))))

(define local-assign
  (lambda (id val env)
    (let ([pair (find id env)])
      (if (not pair)
	  (let ([pair0 (assq id (cdr (user-defined-top-level)))])
	    (if (not pair0)
		(if (top-level-bound? id)
		    (set-top-level-value! id val)
		    (error '() "variable ~s is not bound." id))
		(let ([v (cdr pair0)])
		  (set! set-frames (cons (list 0 id val) set-frames))
		  (for-each (lambda (l)
			      (if display-env?
				  (begin				    
				    (pretty-env
				      (list '!S (output-mgr l)))
				    (newline))))
		    set-frames)
		  (set! set-frames '())
		  (set-car! v val))))
	  (let ([v (cdr pair)]
		[int (find-int id env)])
	    (if (exp-env? (car v))
		(error 'assign "Wrong value in env ~s" v)
		(begin
		  (set! set-frames (cons (list int id val) set-frames))
		  (for-each (lambda (l)
			      (if display-env?
				  (begin
				    (pretty-env
				      (list '!S (output-mgr l)))
				    (newline))))
		    set-frames)
		  (set! set-frames '())
		  (set-car! v val))))))))

(define local-get-max-frame-num (lambda () max-frame-num))
(define local-get-num-of-frames (lambda () num-of-frames))
(define local-get-used-frames (lambda () used-frames))
(define local-get-track-parent (lambda () track-parent))

(set! exist-in-top-level? local-exist-in-top-level?)
(set! add-top-level local-add-top-level)
(set! assign-top-level local-assign-top-level)
(set! get-top-level local-get-top-level)
(set! GC local-GC)
(set! lookup local-lookup)
(set! assign local-assign)
(set! new-env local-new-env)
(set! env-reset local-env-reset)
(set! env-init local-env-init)
(set! set-new&set local-set-new&set)
(set! get-max-frame-num local-get-max-frame-num)
(set! get-num-of-frames local-get-num-of-frames)
(set! get-used-frames local-get-used-frames)
(set! get-track-parent local-get-track-parent)
)

