;;; Copyright (C) 1995 Sho-Huan Simon Tung
;;; See ./VES2/COPYING for details

(define ves #f)

(let ([to-ves-repl #f]
      [quit #f])

;;; This is where program starts.
  (define ves-local   
    (lambda (pnt-graph)
      (call/cc
	(lambda (k)
	  (set! quit k)
	  (call/cc (lambda (k) (set! to-ves-repl
				 (lambda () (k 'ignored)))))
	  (ves-repl pnt-graph)))))

  (define do-reset
    (lambda ()
      (set! current-exp-env #f)
      (set! debug-flag #f)
      (set! run-till '())
      (set! call-cc-arg #f)
      (set! call-cc-called #f)
      (set! new-control #f)
      (set! cont-called #f)
      (set! steps 0)
      (env-reset)))

  (define ves-repl
    (lambda (pnt-graph)
      (parameterize ([print-graph pnt-graph]
		     [error-handler
		       (lambda (symbol msg . objects)
			 (if (symbol? symbol)
			     (begin
			       (newline)
			       (apply printf (string-append "Error in "
					       (symbol->string symbol) ": "
					       msg) objects)
			       (newline))
			     (begin
			       (newline)
			       (apply printf (string-append "Error: "
					       msg) objects)
			       (newline)))
			 (to-ves-repl))])
	(display "-> ")
	(newline)
	(let loop ([command (expand (read))])
	  (if (pair? command)
	      (case (car command)
		[(reset)
		 (pretty-print (list '!Z 0 0))
		 (do-reset)]
		[(debug) (set! debug-flag (not debug-flag))]
		[(GC)
		 (display "Garbage collecting ...")
		 (newline)  ; don't erase - for a front end bug.
		 (let ([free (GC (get-top-level) current-exp-env)])
		   (if (and display-env? (not (null? free)))
			 (for-each (lambda (l)
				     (pretty-print (list '!C l)))
			   free)))
		 (pretty-print (list '!Z 0 0))]
		[(erase-env)
		 (display "Garbage collecting ...")
		 (newline)
		 (let ([free (GC (get-top-level) current-exp-env)])
		   (if (and display-env? (not (null? free)))
		       (for-each (lambda (l)
				   (pretty-print (list '!C l)))
				 free)))
		 (if display-env?
		     (begin
		       ((rec loop
			  (lambda (n)
			    (if (>= n 0)
				(let ([this-frame (vector-ref
						   (get-used-frames)
						   n)])
				  (if (list? this-frame)
				      (begin
					(pretty-print (list '!C n))
					(kill-time (* 1000
						      (length this-frame)))))
				  (loop (- n 1)))))) (get-max-frame-num))
		       (set! display-env? #f)))
		 (pretty-print (list '!Z 0 0))
		 (newline)]
		[(display-env)
  	         (display "Garbage collecting ...")
		 (newline)
		 (let ([free (GC (get-top-level) current-exp-env)])
		   (if (and display-env? (not (null? free)))
		       (for-each (lambda (l)
				   (pretty-print (list '!C l)))
				 free)))
		 (if (not display-env?)
		     (begin
		       ((rec loop
			  (lambda (n)
			    (if (<= n (get-max-frame-num))
				(let ([this-frame (vector-ref
						    (get-used-frames) n)])
				  (if (list? this-frame)
				      (let ([parent-num
					      (vector-ref 
						(get-track-parent) n)])
					(if (= -1 parent-num)
					    (set! parent-num 0))
					(pretty-env
					  (list '!N
						(output-mgr
						  (cons (car this-frame)
  						    (cons parent-num
						      (cdr this-frame))))))
					(newline)))
				  (loop (+ n 1)))))) 0)
		       (set! display-env? #t)))
		 (pretty-print (list '!Z 0 0))
		 (newline)]
		[(display-env-tree)
		 (display "Drawing ...")
		 (newline)
		 (count-width (get-max-frame-num))]
		[(change-value)
		 (let ([frame-num (cadr command)]
		       [var (caddr command)]
		       [new-exp (cadddr command)]
		       [used-frames (get-used-frames)]
		       [parent-frames (get-track-parent)])
		   (if (not (and (number? frame-num)
			    (not (eq? 0 (vector-ref used-frames frame-num)))))
		       (error 'change-value "Invalid frame number ~s"
			 frame-num))
		   (if (not (symbol? var))
		       (error 'change-value "Invalid identifier ~s" var))
		   (let ([frames  ((rec loop
				     (lambda (f)
				       (if (= f 0)
					   (list)
					   (cons
					     (vector-ref used-frames f)
					     (loop
					       (vector-ref
						 parent-frames f))))))
				   frame-num)])
		     (fluid-let ([steps -1])
		       (let ([new-v-env (meaning (make-exp-env new-exp frames
						   #f 0))])
			 (assign var (exp-env-exp new-v-env) frames)))))
		 (pretty-print (list '!Z 0 0))
		 (newline)]

		[(v view)
		 (let ([free (GC (get-top-level) '())])
		   (if (and display-env? (not (null? free)))
		       (for-each (lambda (l)
				   (pretty-print (list '!C l)))
			 free)))
		 (set-new&set)
		 (set! current-exp-env
		   (make-exp-env (cadr command)
				 (list (list 0)) (tail-enable?) 0))
		 (set! steps 0)
		 (my-pretty (output-mgr current-exp-env))]
		[(define-primitive)
		 (set-top-level-value! (cadr command) (eval (caddr command)))]
		[(load) (my-pretty (output-mgr (do-load command)))]
		[(sg show-global)
		 (printf "UserGlobal =~%")
		 (parameterize ([print-graph #f])
		   (pretty-print (get-top-level)))]
		[(micro)
		 (set! micro (not micro))
		 (if micro
		     (display "Trace in detailed mode.")
		     (display "Trace in normal mode."))
		 (newline)]
		[(run-till)
		 (if (and (pair? (cdr command)) (symbol? (cadr command)))
		     (begin
		       (if (eq? 'call/cc (cadr command))
			   (set! break-on-call/cc #t)
			   (set! run-till (cons (cadr command) run-till)))
		       (printf "Set break point: ~s.~%" (cadr command))))]
		[(no-run-till)
		 (if (and (pair? (cdr command)) (symbol? (cadr command)))
		     (if (eq? 'call/cc (cadr command))
			 (set! break-on-call/cc #f)
			 (begin
			   (set! run-till (remq (cadr command) run-till))
			   (printf
			     "Clear break point: ~s.~%" (cadr command)))))]
		[(f forward)
		 (pretty-print (list '!Z 0 0))
		 (set-new&set)
		 (set! call-cc-called #f)
		 (set! cont-called #f)
		 (if (and (pair? (cdr command)) (integer? (cadr command)))
		     (set! steps (cadr command))
		     (error "Does not recognize command: ~s" command))
		 ((rec loop
		    (lambda (crnt-exp-env)
		      (if (and (exp-env? crnt-exp-env)
			       (value? (exp-env-env crnt-exp-env))
			       (not (eq? 'to-receive
					 (exp-env-env crnt-exp-env))))
			  (begin
			    (set! current-exp-env crnt-exp-env)
			    (my-pretty (output-mgr  crnt-exp-env)))
			  (if (= 0 steps)
			      (if call-cc-called
				  (begin
				    (set! call-cc-called #f)
				    (let* ([c-obj (make-exp-env
						    `(!K ,crnt-exp-env)
						    'cont-obj
						    (tail-enable?) 0)]
					   [c-app-obj `(,call-cc-arg ,c-obj)])
				      (set! current-exp-env
					(subst c-app-obj cont-val crnt-exp-env)))
				    (set-new&set)
				    (set! cont-called #f)
				    (if break-on-call/cc
					(my-pretty (output-mgr current-exp-env))
					(begin
					  (set! steps (cadr command))
					  (loop (meaning current-exp-env)))))
					  
				  (if cont-called
				      (begin
					(set! cont-called #f)
					(set! call/cc-called #f)
					(set-new&set)
					(set! current-exp-env new-control)
					(if break-on-call/cc
					    (my-pretty (output-mgr
							 current-exp-env))
					    (begin
					      (set! steps (cadr command))
					      (loop (meaning current-exp-env)))))
				      (begin
					(set! current-exp-env crnt-exp-env)
					(my-pretty (output-mgr
						     crnt-exp-env)))))
			      (loop (meaning crnt-exp-env))))))
		  current-exp-env)]
		[(quit exit)
		 (do-reset)
		 (quit #f)]
		[else (error 'ves-repl "Bad command ~s" command)])
	      (error 'ves-repl "Bad command ~s" command))
	  (display "-> ")
	  (newline)
	  (loop (expand (read)))))))

  (set! ves ves-local)
  )

(define count-width
  (lambda (max)
    (let ([child-v (make-vector (+ max 1) '())]
	  [child-width (make-vector (+ max 1) 1)]
	  [used-frames (get-used-frames)]
	  [parent (get-track-parent)]
	  [pos-v (make-vector (+ max 1) '())])
      (define count-child
	(lambda (max)    ; use track-parent to build a vector of child list.
	    ((rec loop
	       (lambda (n)
		 (if (> n max)
		     child-v
		     (begin
		       (let loop1 ([m n])
			 (if (<= m max)
			     (begin
			       (if (= n (vector-ref parent m))
				   (vector-set! child-v n
				     (cons m (vector-ref child-v n))))
			       (loop1 (+ m 1)))))
		       (loop (+ n 1)))))) 0)))

      (define cal-depth
	(lambda ()
	  (let ([max-d 0])
	    ((rec loop0
		  (lambda (j)
		    (if (<= j max)
			(begin
			  ((rec loop
				(lambda (i d)
				  (if (not (= -1 (vector-ref parent i)))
				      (loop (vector-ref parent i) (+ d 1))
				      (if (> d max-d)
					  (set! max-d d))))) j 1)
			  (loop0 (+ j 1)))
			max-d))) 0))))

      (define cal-width
	(lambda (max)
	  ((rec loop
	     (lambda (n)
	       (if (< n 0)
		   child-width
		   (begin
		     (if (not (null? (vector-ref child-v n)))
			 (vector-set! child-width n
			   (accu (vector-ref child-v n))))
		     (loop (- n 1)))))) max)))

      (define accu
	(lambda (l)
	  (if (null? l)
	      0
	      (+ (vector-ref child-width (car l)) (accu (cdr l))))))

      (define make-chi-alloc
	(lambda (l)
	  (if (null? l)
	      '()
	      (cons (vector-ref child-width (car l))
		    (make-chi-alloc (cdr l))))))

      (define pre-order ; not done
	(lambda (node-num parent where level)
	  (vector-set! pos-v node-num (list where level))
	  (pretty-env
	    (list '!T (list node-num parent where level
	      (output-mgr
		(let ([f (vector-ref used-frames node-num)])
		  (if (= 0 node-num)
		      (reverse (cdr f))
		      (cdr f)))))))
	  (newline)
	  (if (not (and (= 0 parent) (= 0 node-num)))
	      (pretty-print (list '!A 
;			      (append (vector-ref pos-v parent))
			      (list node-num parent where level))))
	  (let ([chi (reverse (vector-ref child-v node-num))]
		[chi-wid (vector-ref child-width node-num)])
	    (let ([chi-pos (compute-chi-pos (make-chi-alloc chi) where)])
	      ((rec loop
		 (lambda (chi chi-pos)
		   (if (not (null? chi))
		       (begin
			 (pre-order (car chi) node-num (car chi-pos)
			   (+ 1 level))
			 (loop (cdr chi) (cdr chi-pos)))))) chi chi-pos)))))

      (define compute-chi-pos
	(lambda (chi-wid where)
	  (let ([l (length chi-wid)]
		[pos '*])
	    (let ([zero (if (odd? l) '(0) '())]
		  [negative
		    (reverse
		      ((rec loop
			 (lambda (n chi-wid)
				 (if (> n (/ l 2))
				     (begin
				       (set! pos (if (odd? l)
						     (cdr chi-wid)
						     chi-wid))
				       '())
				     (cons (car chi-wid)
					   (loop (+ n 1) (cdr chi-wid))))))
			     1 chi-wid))])
		  (let ([new-neg
			  (reverse
			    ((rec loop
				    (lambda (l so-far)
				      (if (null? l)
					  '()
					  (cons
					    (+ so-far
					       (floor
						 (/
						   (- 0
						      (if (= 1 (car l))
							  2
							  (car l))) 2)))
					    (loop (cdr l)
						  (+ so-far
						     (- 0 (car l))))))))
				  negative 0))]
			[new-pos ((rec loop
				    (lambda (l so-far)
				      (if (null? l)
					  '()
					  (cons
					    (+ so-far (ceiling
							(/
							  (if (= 1 (car l))
							      2
							      (car l)) 2)))
					    (loop (cdr l)
						  (+ so-far (car l)))))))
				  pos 0)])
	      (map (lambda (n) (+ where n))
		   (if (null? zero)
		       (append (map (lambda (i) (+ i 1)) new-neg) new-pos)
		       (append new-neg (append zero new-pos)))))))))

      (count-child max)
      (cal-width max)
      (pretty-print (list '!Z (vector-ref child-width 0) (cal-depth)))
      (pre-order 0 0 0 0))))
	
